﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UDITA;

namespace IntRedBlackTree
{
    public class IntRedBlackTree
    {
        public class Node
        {
            public int key;
            public Object value;
            public Node parent;
            public Node left, right;
            public bool color;

            public static Node generate(Node parent, int min, int max)
            {
                Node node = new Node();
                node.parent = parent;
                node.key = Verify.getInt(min, max);
                node.color = Verify.getBoolean();
                return node;
            }
        }

        public Node root;

        public const bool RED = false;

        public const bool BLACK = true;

        public const int FALSE = -1;

        public static int treeSize;
        public static int maxNodeValue;

        public int size;

        ////************************** Generators *****************************

        public void generateRedBlackTree()
        {
            generateSortedTree();
            bool property = repOkColors();
            Assume.assume(property);
        }

        private void generateSortedTree()
        {
            size = treeSize;
            root = getSubTree(null, size, 0, maxNodeValue);
        }

        private static Node getSubTree(Node parent, int size, int min, int max)
        {
            if (size == 0)
                return null;
            Node next = Node.generate(parent, min, max);
            if (size == 1)
                return next;

            int leftSize = next.key - min;
            int rightSize = size - 1 - leftSize;

            next.left = getSubTree(next, leftSize, min, next.key - 1);
            next.right = getSubTree(next, rightSize, next.key + 1, max);

            return next;
        }

        private bool repOkColors()
        {
            return repOkColorsRec(root) != FALSE;
        }


        ////*************************** Filters *******************************

        // with ObjectPool
        public void generateGraph()
        {
            size = treeSize;
            ObjectPool<Node> pool = new ObjectPool<Node>(typeof(Node), treeSize, true);

            root = pool.getNew();
            foreach (Node n in pool)
            {
                if (n != null)
                {
                    n.color = Verify.getBoolean();
                    n.key = Verify.getInt(0, maxNodeValue);
                    n.left = pool.getAny();
                    n.right = pool.getAny();
                    n.parent = pool.getAny();
                }
            }
        }

        //******************************* repOK ********************************* 

        public bool repOK()
        {
            if (root == null)
                return size == 0;
            // RootHasNoParent
            if (root.parent != null)
                return debug("RootHasNoParent");
            HashSet<Node> visited = new HashSet<Node>();
            visited.Add(root);
            LinkedList<Node> workList = new LinkedList<Node>();
            workList.AddLast(root);
            while (workList.Count > 0)
            {
                Node current = workList.First();
                workList.RemoveFirst();

                Node cl = current.left;
                if (cl != null)
                {
                    if (!visited.Add(cl))
                        return debug("Acyclic");
                    if (cl.parent != current)
                        return debug("parent_Input1");
                    workList.AddLast(cl);
                }
                Node cr = current.right;
                if (cr != null)
                {
                    if (!visited.Add(cr))
                        return debug("Acyclic");
                    if (cr.parent != current)
                        return debug("parent_Input2");
                    workList.AddLast(cr);
                }
            }
            // SizeOk
            if (visited.Count != size)
                return debug("SizeOk");

            if (repOkColorsRec(root) == FALSE)
                return false;
            return repOkKeysAndValues();
        }

        private int repOkColorsRec(Node node)
        {
            if (node == null)
                return 0;

            // red has only black children
            if (node.color == RED)
            {
                if (node.left != null && node.left.color == RED)
                    return FALSE;
                if (node.right != null && node.right.color == RED)
                    return FALSE;
            }

            // simple path from root to null have the same number of black nodes
            int leftNumber = repOkColorsRec(node.left);
            if (leftNumber == FALSE)
                return FALSE;

            int rightNumber = repOkColorsRec(node.right);
            if (rightNumber == FALSE)
                return FALSE;

            if (leftNumber != rightNumber)
                return FALSE;

            return (leftNumber + (node.color == BLACK ? 1 : 0));
        }

        private bool repOkKeysAndValues()
        {
            // way to determine if the keys are ordered
            if (!orderedKeys(root, null, null))
                return debug("BST");
            // touch values
            LinkedList<Node> workList = new LinkedList<Node>();
            workList.AddLast(root);
            while (workList.Count > 0)
            {
                Node current = workList.First();
                workList.RemoveFirst();

                if (current.left != null)
                    workList.AddLast(current.left);
                if (current.right != null)
                    workList.AddLast(current.right);
            }
            return true;
        }

        private bool orderedKeys(Node e, object min, object max)
        {
            if (((min != null) && (e.key.CompareTo(min) <= 0))
                || ((max != null) && (e.key.CompareTo(max) >= 0)))
                return false;
            if (e.left != null)
                if (!orderedKeys(e.left, min, e.key))
                    return false;
            if (e.right != null)
                if (!orderedKeys(e.right, e.key, max))
                    return false;
            return true;
        }

        private bool debug(string s)
        {
            //Console.Out.WriteLine(s);
            return false;
        }
        //*******************************************************************************************

        protected Node parent(Node n)
        {
            return n == null ? null : n.parent;
        }

        protected Node left(Node n)
        {
            return n == null ? null : n.left;
        }

        protected Node right(Node n)
        {
            return n == null ? null : n.right;
        }
        
        public bool containsKey(int key)
        {
            return findNode(key) != null;
        }

        public Object get(int key)
        {
            Node x = findNode(key);
            return x == null ? null : x.value;
        }

        public void put(int key, Object value)
        {
            Node x = new Node();
            x.key = key;
            x.value = value;
            treeInsert(x);
            size++;
        }

        public Object remove(int key)
        {
            Node ret = treeDelete(findNode(key));
            if (ret == null)
                return null;
            size--;
            return ret.value;
        }

        protected bool getColor(Node x)
        {
            // null nodes are colored black
            return x == null ? BLACK : x.color;
        }

        protected void setColor(Node x, bool color)
        {
            if (x != null)
            {
                x.color = color;
            }
        }

        protected Node findNode(int key)
        {
            Node cur = root;
            while (cur != null && key.CompareTo(cur.key) != 0)
                if (key.CompareTo(cur.key) < 0)
                    cur = cur.left;
                else
                    cur = cur.right;
            return cur;
        }

        protected void leftRotate(Node x)
        {
            Node y = x.right;
            x.right = y.left;
            if (y.left != null)
                y.left.parent = x;
            y.parent = x.parent;
            if (x.parent == null)
                root = y;
            else if (x == x.parent.left)
                x.parent.left = y;
            else
                x.parent.right = y;
            y.left = x;
            x.parent = y;
        }

        protected virtual void rightRotate(Node x)
        {
            Node y = x.left;
            x.left = y.right;
            if (y.right != null)
                y.right.parent = x;
            y.parent = x.parent;
            if (x.parent == null)
                root = y;
            else if (x == x.parent.right)
                x.parent.right = y;
            else
                x.parent.left = y;
            y.right = x;
            x.parent = y;
        }

        protected virtual void treeInsert(Node z)
        {
            Node y = null;
            Node x = root;
            while (x != null)
            {
                y = x;
                if (z.key.CompareTo(x.key) < 0)
                    x = x.left;
                else if (z.key.CompareTo(x.key) == 0)
                {
                    x.value = z.value;
                    size--;
                    return;
                }
                else
                    x = x.right;
            }
            z.parent = y;
            if (y == null)
                root = z;
            else if (z.key.CompareTo(y.key) < 0)
                y.left = z;
            else
                y.right = z;
            z.left = null;
            z.right = null;
            z.color = RED;

            treeInsertFix(z);
        }

        protected virtual void treeInsertFix(Node z)
        {
            while (getColor(z.parent) == RED)
            {
                if (parent(z) == left(parent(parent(z))))
                {
                    Node y = right(parent(parent(z)));
                    if (getColor(y) == RED)
                    {
                        setColor(parent(z), BLACK);
                        setColor(y, BLACK);
                        setColor(parent(parent(z)), RED);
                        z = parent(parent(z));
                    }
                    else
                    {
                        if (z == right(parent(z)))
                        {
                            z = parent(z);
                            leftRotate(z);
                        }
                        setColor(parent(z), BLACK);
                        setColor(parent(parent(z)), RED);
                        if (parent(parent(z)) != null)
                            rightRotate(parent(parent(z)));
                    }
                }
                else
                {
                    Node y = left(parent(parent(z)));
                    if (getColor(y) == RED)
                    {
                        setColor(parent(z), BLACK);
                        setColor(y, BLACK);
                        setColor(parent(parent(z)), RED);
                        z = parent(parent(z));
                    }
                    else
                    {
                        if (z == left(parent(z)))
                        {
                            z = parent(z);
                            rightRotate(z);
                        }
                        setColor(parent(z), BLACK);
                        setColor(parent(parent(z)), RED);
                        if (parent(parent(z)) != null)
                            leftRotate(parent(parent(z)));
                    }
                }
            }
            root.color = BLACK;
        }


        protected virtual Node treeDelete(Node z)
        {
            if (z == null)
                return null;
            Node x, y;
            if (z.left == null || z.right == null)
                y = z;
            else
                y = getIOS(z);
            if (y.left != null)
                x = y.left;
            else
                x = y.right;
            if (x != null)
                x.parent = y.parent;
            if (y.parent == null)
                root = x;
            else if (y == y.parent.left)
                y.parent.left = x;
            else
                y.parent.right = x;
            if (y != z)
            {
                z.key = y.key;
                z.value = y.value;
            }
            if (getColor(y) == BLACK)
                if (x == null)
                    // hack
                    treeDeleteFix(y);
                else
                    treeDeleteFix(x);
            return y;
        }

        protected void treeDeleteFix(Node x)
        {
            while (x.parent != null && getColor(x) == BLACK)
            {
                if (x == x.parent.left || x.parent.left == null)
                {
                    Node w = x.parent.right;
                    if (w == null)
                        return;
                    if (getColor(w) == RED)
                    {
                        w.color = BLACK;
                        x.parent.color = RED;
                        leftRotate(x.parent);
                        w = x.parent.right;
                    }
                    if (getColor(w.left) == BLACK && getColor(w.right) == BLACK)
                    {
                        w.color = RED;
                        x = x.parent;
                    }
                    else
                    {
                        if (getColor(w.right) == BLACK)
                        {
                            w.left.color = BLACK;
                            rightRotate(w);
                            w = x.parent.right;
                        }
                        w.color = x.parent.color;
                        x.parent.color = BLACK;
                        if (w.right != null)
                            w.right.color = BLACK;
                        leftRotate(x.parent);
                        x = root;
                    }
                }
                else
                {
                    Node w = x.parent.left;
                    if (w == null)
                        return;
                    if (getColor(w) == RED)
                    {
                        w.color = BLACK;
                        x.parent.color = RED;
                        rightRotate(x.parent);
                        w = x.parent.left;
                    }
                    if (getColor(w.right) == BLACK && getColor(w.left) == BLACK)
                    {
                        w.color = RED;
                        x = x.parent;
                    }
                    else
                    {
                        if (getColor(w.left) == BLACK)
                        {
                            w.right.color = BLACK;
                            leftRotate(w);
                            w = x.parent.left;
                        }
                        w.color = x.parent.color;
                        x.parent.color = BLACK;
                        if (w.left != null)
                            w.left.color = BLACK;
                        rightRotate(x.parent);
                        x = root;
                    }
                }
            }
            x.color = BLACK;
        }

        protected Node getIOS(Node z)
        {
            Node x, y = null;
            x = z.right;
            while (x != null)
            {
                y = x;
                x = x.left;
            }
            return y;
        }

    }
}
