using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Pex.Framework;

namespace UDITA
{
    public sealed class ObjectPool<T> : IEnumerable<T>
    {
        // to indicate that pool is infinite
        private const int INFINITE_FLAG = -1;

        // keeps objects
        internal List<Object> objects;

        // is pool infinite
        private bool infinite;

        // size of the pool (does have sense for finite pools)
        private int size;

        // current position in object pool
        private int index;

        // is "null" included
        private bool includeNull;

        // type of objects in the pool
        private Type clz;

        /* ********************** */

        /**
         * Creates infinite Object Pool, "null" is included
         * 
         * @param clz
         *          type of objects in the pool
         * 
         */
        public ObjectPool(Type clz) : this(clz, INFINITE_FLAG, true) {}

        /**   
         * Creates finite Object Pool of the given size, "null" is included
         * 
         * @param clz
         *          type of objects in the pool
         * @param size
         *          max objects in pool
         */

        public ObjectPool(Type clz, int size) : this(clz, size, true) {}

        /**
         * Creates infinite Object Pool.
         * 
         * @param clz
         *          type of objects in the pool
         * @param includeNull
         *          true if "null" is included in the pool, false otherwise
         */
        public ObjectPool(Type clz, bool includeNull) : this(clz, INFINITE_FLAG, includeNull) {}

        /**
         * Creates finite Object Pool.
         * 
         * @param clz
         *          type of objects in the pool
         * @param size
         *          max number of objects in the pool
         * @param includeNull
         *          true if "null" is included, false otherwise
         */
        public ObjectPool(Type clz, int size, bool includeNull)
        {
            this.clz = clz;
            this.infinite = size == INFINITE_FLAG;
            this.size = size + (includeNull ? 1 : 0);
            this.includeNull = includeNull;
            this.index = (includeNull ? 0 : -1);
            Init();
        }

        private void Init()
        {
            objects = new List<Object>();
            if (includeNull)
                objects.Add(null);

            if (!infinite)
            {
                // create objects
                for (int i = includeNull ? 1 : 0; i < size; i++)
                {
                    try
                    {
                        objects.Add(clz.GetConstructor(new Type[0]).Invoke(new object[0]));
                    }
                    catch (Exception)
                    {
                    }
                }
            }
        }

        // used just in case of infinite pool
        private void ExpandPool()
        {
            try
            {
                objects.Add(clz.GetConstructor(new Type[0]).Invoke(new object[0]));
            }
            catch (Exception)
            {
            }
        }

        private T GetObject(int ix)
        {
            if (infinite && index == ix)
            {
                ExpandPool();
            }
            return (T)objects.ElementAt<object>(ix);
        }

        /**
         * Returns objects from the pool. Result of this method will take all objects
         * from the pool assigned so far and new one if that is possible.
         * 
         * @return All objects from the pool assigned so far and new one if it
         *         possible
         */
        public T GetAny()
        {  
            PexAssume.IsFalse(!infinite && size == 0);

            int ix;
            if (infinite)
            {
                ix = PexChoose.DefaultSession.ChooseValueFromRange("ix", 0, index + 2);

            }
            else
            {
                ix = PexChoose.DefaultSession.ChooseValueFromRange("ix", 0, Math.Min(index + 2, size));
            }

            if (ix == index + 1)
            {
                index++;
            }
            return GetObject(ix);
        }

        /**
         * Returns new object from the pool if new one can be created,
         * "null" will never be returned.
         * 
         * @return new object from the pool if that is possible.
         */
        public T GetNew()
        {
            PexAssume.IsFalse(!infinite && (index + 1 == size || size == 0));
            return GetObject(++index);
        }

        IEnumerator<T> IEnumerable<T>.GetEnumerator()
        {
            return new ObjectPoolEnumerator<T>(this);
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return new ObjectPoolEnumerator<T>(this);
        }
    }

    public class ObjectPoolEnumerator<T> : IEnumerator<T>
    {
        private ObjectPool<T> op;
        private int next = 0;

        public ObjectPoolEnumerator(ObjectPool<T> op)
        {
            this.op = op;
        }

        T IEnumerator<T>.Current
        {
            get { return (T)op.objects.ElementAt<object>(next); }
        }

        object System.Collections.IEnumerator.Current
        {
            get { return (T)op.objects.ElementAt<object>(next); }
        }

        bool System.Collections.IEnumerator.MoveNext()
        {
            next++;
            return (next < op.objects.Count);
        }

        void System.Collections.IEnumerator.Reset()
        {
            next = 0;
        }

        void IDisposable.Dispose() {}

    }
}