package intredblacktree;

import gov.nasa.jpf.delayed.ObjectPool;
import gov.nasa.jpf.jvm.Verify;
import gov.nasa.jpf.jvm.generators.StructureCounter;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;

public class IntRedBlackTreeUDITA extends IntRedBlackTree {
  protected static final int FALSE = -1;

  private static ObjectPool<Node> entries;

  public static class NodeUDITA extends Node {
    static Node generate(Node parent, int min, int max) {
      Node node = new Node();
      node.parent = parent;
      node.key = Verify.getInt(min, max);
      node.color = Verify.getBoolean();
      return node;
    }
  }

  // ------------------------ repOK ---------------------------
  public boolean repOK() {
    if (root == null)
      return size == 0;
    if (root.parent != null)
      return debug("RootHasNoParent");
    Set<Node> visited = new java.util.HashSet<Node>();
    visited.add(root);
    LinkedList<Node> workList = new LinkedList<Node>();
    workList.add(root);
    while (!workList.isEmpty()) {
      Node current = (Node) workList.removeFirst();
      Node cl = current.left;
      if (cl != null) {
        if (!visited.add(cl))
          return debug("Acyclic");
        if (cl.parent != current)
          return debug("parent_Input1");
        workList.add(cl);
      }
      Node cr = current.right;
      if (cr != null) {
        if (!visited.add(cr))
          return debug("Acyclic");
        if (cr.parent != current)
          return debug("parent_Input2");
        workList.add(cr);
      }
    }
    if (visited.size() != size)
      return debug("SizeOk");

    if (repOkColorsRec(root) == FALSE)
      return false;
    return repOkKeysAndValues();
  }

  /**
   * This method is copied from Korat repo.
   * 
   * @return true if color property is satisfied, false otherwise
   */
  private int repOkColorsRec(Node node) {
    if (node == null)
      return 0;

    // RedHasOnlyBlackChildren
    if (node.color == RED) {
      if (node.left != null && node.left.color == RED)
        return FALSE;
      if (node.right != null && node.right.color == RED)
        return FALSE;
    }

    // SimplePathsFromRootToNILHaveSameNumberOfBlackNodes
    int leftNumber = repOkColorsRec(node.left);
    if (leftNumber == FALSE)
      return FALSE;

    int rightNumber = repOkColorsRec(node.right);
    if (rightNumber == FALSE)
      return FALSE;

    if (leftNumber != rightNumber)
      return FALSE;

    return (leftNumber + (node.color == BLACK ? 1 : 0));
  }

  private boolean repOkColors() {
    return repOkColorsRec(root) != FALSE;
  }

  private boolean repOkKeysAndValues() {
    if (!orderedKeys(root, null, null))
      return debug("BST");
    LinkedList<Node> workList = new LinkedList<Node>();
    workList.add(root);
    while (!workList.isEmpty()) {
      Node current = (Node) workList.removeFirst();

      if (current.left != null)
        workList.add(current.left);
      if (current.right != null)
        workList.add(current.right);
    }
    return true;
  }

  private boolean orderedKeys(Node e, Integer min, Integer max) {
    if (((min != null) && (compare(e.key, min) <= 0))
        || ((max != null) && (compare(e.key, max) >= 0)))
      return false;
    if (e.left != null)
      if (!orderedKeys(e.left, min, e.key))
        return false;
    if (e.right != null)
      if (!orderedKeys(e.right, e.key, max))
        return false;
    return true;
  }

  private final boolean debug(String s) {
    // System.out.println(s);
    return false;
  }

  protected int compare(Integer k1, Integer k2) {
    return k1.compareTo(k2);
  }

  // ------------------------ finitization ---------------------------
  private static IntRedBlackTreeUDITA finitization(Arguments arguments) {
    IntRedBlackTreeUDITA rbt = arguments.newIntRedBlackTree();
    rbt.size = arguments.generateUpToSize ? Verify.getInt(1, arguments.treeSize) : arguments.treeSize;
    int maxNodeValue = rbt.size - 1;

    entries = new ObjectPool<Node>(Node.class, rbt.size, true);

    rbt.root = entries.getNew();

    for (Node n : entries) {
      if (n != null) {
        n.color = Verify.getBoolean();
        n.key = Verify.getInt(0, maxNodeValue);
        n.left = entries.getAny();
        n.right = entries.getAny();
        n.parent = entries.getAny();
      }
    }
    return rbt;
  }

  // ------------------------ filtering generator ---------------------------
  public static IntRedBlackTreeUDITA generateRedBlackTreeFiltering(
      Arguments arguments) {
    IntRedBlackTreeUDITA rbt = finitization(arguments);
    boolean property = rbt.repOK();
    StructureCounter.incCounters(property);
    Verify.ignoreIf(!property);
    return rbt;
  }

  // ------------------------ combined generator ---------------------------
  public static IntRedBlackTreeUDITA generateRedBlackTreeCombined(
      Arguments arguments) {
    IntRedBlackTreeUDITA rbt = generateSortedTree(arguments);
    boolean property = rbt.repOkColors();
    StructureCounter.incCounters(property);
    Verify.ignoreIf(!property);
    return rbt;
  }

  private static IntRedBlackTreeUDITA generateSortedTree(Arguments arguments) {
    IntRedBlackTreeUDITA st = arguments.newIntRedBlackTree();
    st.size = arguments.generateUpToSize ? Verify.getInt(1, arguments.treeSize)
        : arguments.treeSize;
    int maxNodeValue = st.size - 1;
    st.root = getSubTree(null, st.size, 0, maxNodeValue);
    return st;
  }

  private static Node getSubTree(Node parent, int size, int min, int max) {
    if (size == 0)
      return null;
    Node next = NodeUDITA.generate(parent, min, max);
    if (size == 1)
      return next;

    int leftSize = next.key - min;
    int rightSize = size - 1 - leftSize;

    next.left = getSubTree(next, leftSize, min, next.key - 1);
    next.right = getSubTree(next, rightSize, next.key + 1, max);

    return next;
  }

  // ------------------------ main ---------------------------
  public static void superMain(String[] args,
      IntRedBlackTreeUDITA toBeUsedAsFactory) {
    Arguments arguments = Arguments.createAndCheckArguments(args,
        toBeUsedAsFactory);

    IntRedBlackTreeUDITA rbt;
    if (!arguments.isCombined)
      rbt = generateRedBlackTreeFiltering(arguments);
    else
      rbt = generateRedBlackTreeCombined(arguments);

    if (arguments.method.equals(Arguments.REMOVE_METHOD)) {
      int key = Verify.getInt(0, arguments.treeSize);
      rbt.remove(key);
      assert (rbt.repOK());
    } else if (arguments.method.equals(Arguments.PUT_METHOD)) {
      int key = Verify.getInt(-1, arguments.treeSize + 1);
      rbt.put(key, null);
      assert (rbt.repOK());
    }

    if (arguments.print)
      System.out.println(rbt);
  }

  public static void main(String[] args) {
    superMain(args, new IntRedBlackTreeUDITA());
  }

  public IntRedBlackTreeUDITA newIntRedBlackTree() {
    return new IntRedBlackTreeUDITA();
  }

  public static class Arguments {
    public static final String printSwitch = "-p";
    public static final String sizeSwitch = "-s";
    public static final String filteringModeSwitch = "-f";
    public static final String combinedModeSwitch = "-k";
    public static final String upToSizeSwitch = "-up";
    public static final String methodSwitch = "-()";

    public static final String PUT_METHOD = "put";
    public static final String REMOVE_METHOD = "remove";

    final IntRedBlackTreeUDITA rbt;
    final int treeSize;
    final boolean generateUpToSize;
    final boolean isCombined;
    final boolean print;
    final public String method;

    public Arguments(IntRedBlackTreeUDITA rbt, int treeSize,
        boolean generateUpToSize, boolean isCombined, boolean print,
        String method) {
      this.rbt = rbt;
      this.treeSize = treeSize;
      this.generateUpToSize = generateUpToSize;
      this.isCombined = isCombined;
      this.print = print;
      this.method = method;
    }

    public IntRedBlackTreeUDITA newIntRedBlackTree() {
      return rbt.newIntRedBlackTree();
    }

    private static boolean checkModeAndReturn(List<String> args) {
      if ((args.contains(filteringModeSwitch) && args
          .contains(combinedModeSwitch))
          || (!args.contains(filteringModeSwitch) && !args
              .contains(combinedModeSwitch)))
        printHelpAndExit("incorrect mode");
      return args.contains(combinedModeSwitch);
    }

    private static int checkSizeAndReturn(List<String> args) {
      if (!args.contains(sizeSwitch))
        printHelpAndExit("size must be specified");
      return Integer.parseInt(args.get(args.indexOf(sizeSwitch) + 1));
    }

    private static String checkMethodAndReturn(List<String> args) {
      if (!args.contains(methodSwitch))
        printHelpAndExit("method must be specified");
      String method = args.get(args.indexOf(methodSwitch) + 1);
      if (!method.equals(PUT_METHOD) && !method.equals(REMOVE_METHOD))
        printHelpAndExit("incorrect method");
      return method;
    }

    private static Arguments createAndCheckArguments(String[] args,
        IntRedBlackTreeUDITA rbt) {
      List<String> listOfArgs = new LinkedList<String>();
      for (String s : args)
        listOfArgs.add(s);

      int treeSize = checkSizeAndReturn(listOfArgs);
      boolean generateUpToSize = listOfArgs.contains(upToSizeSwitch);
      boolean isCombined = checkModeAndReturn(listOfArgs);
      boolean print = listOfArgs.contains(printSwitch);
      String method = checkMethodAndReturn(listOfArgs);

      return new Arguments(rbt, treeSize, generateUpToSize, isCombined, print,
          method);
    }

    private static void printHelpAndExit(String info) {
      printHelp();
      throw new RuntimeException("incorrect arguments! " + info);
    }

    private static void printHelp() {
      System.out.println("IntRedBlackTree[BX] -s size {-k|-f} [-p] [-up] -() method");
    }
  }

}
